// for the timer class
#include "profiling.h"

#include <vector>
#include <set>
#include <unordered_set>
#include <iostream>
#include <random>
#include <cstdint>

using namespace std;

// This code is for generating random numbers
int32_t RandomInt() {
  static minstd_rand rng;
  return rng();
}

volatile int poison;

// The important code

void VectorBenchmark(size_t n_elements) {
  vector<int32_t> v;
  Timer timer;
  // Insert n_elements
  for (size_t i = 0; i < n_elements; ++i) {
    v.push_back(RandomInt());
  }
  timer.Tock();
  cout << "push_back takes " << timer.ElapsedTime() << " ms\n";

  // Create a second vector to prevent cache-related effects
  vector<int32_t> v2;
  timer.Reset();
  v2.reserve(n_elements);
  for (size_t i = 0; i < n_elements; ++i) {
    v2.push_back(RandomInt());
  }
  timer.Tock();
  cout << "reserve+push_back takes " << timer.ElapsedTime() << " ms\n";

  // Pop all elements one-by-one
  timer.Reset();
  timer.Tick();
  for (size_t i = 0; i < n_elements; ++i) {
    v2.pop_back();
    // do something nontrivial to prevent the compiler from optimizing
    // this loop
    poison = v2.size();
  }
  timer.Tock();
  cout << "pop_back takes " << timer.ElapsedTime() << " ms\n";

  // Access all elements sequentially and do something
  int32_t sum = 0;
  timer.Reset();
  for (size_t i = 0; i < n_elements; ++i) {
    sum += v[i];
  }
  timer.Tock();
  cout << "[] takes " << timer.ElapsedTime() << " ms, btw the sum is " << sum << "\n";

  // Access all elements sequentially using .at() and do something
  sum = 0;
  timer.Reset();
  for (size_t i = 0; i < n_elements; ++i) {
    sum += v.at(i);
  }
  timer.Tock();
  cout << ".at() takes " << timer.ElapsedTime() << " ms, btw the sum is " << sum << "\n";

  // Access all elements sequentially using iterators/for-range
  sum = 0;
  timer.Reset();
  for (const auto& x : v) {
    sum += x;
  }
  timer.Tock();
  cout << "range-for takes " << timer.ElapsedTime() << " ms, btw the sum is " << sum << "\n";

  // Access random elements
  sum = 0;
  timer.Reset();
  timer.Tick();
  for (size_t n_times = n_elements; n_times > 0; --n_times) {
    size_t i = RandomInt() % n_elements;
    sum += v[i];
  }
  timer.Tock();
  cout << "random access takes " << timer.ElapsedTime() << " ms, btw the sum is " << sum << "\n";
}

// The main driver for the benchmark
void RunBenchmark(size_t n_elements) {
  cout << "N = " << n_elements << "\n\n";
  // Initialize the rng
  RandomInt();
  // Measure how long it takes to generate n_elements as a baseline
  Timer timer;
  for (size_t i = 0; i < n_elements; ++i) {
    RandomInt();
  }
  timer.Tock();
  cout << "it takes " << timer.ElapsedTime() << " ms to generate " << n_elements << " elements\n\n";
  size_t garbage = 0;
  timer.Reset();
  for (size_t i = 0; i < n_elements; ++i) {
    garbage += RandomInt() % n_elements;
  }
  timer.Tock();
  cout << "it takes " << timer.ElapsedTime() << " ms to generate " << n_elements << " indices. Here is some garbage: " << garbage << "\n\n";
  cout << "Vector benchmark:\n\n";
  VectorBenchmark(n_elements);
}

// The code below is the harness doing argument parsing etc. It is not
// important for this class.

void Usage(const std::string& name) {
  cerr << "Usage: " << name << " number-of-elements\n";
}

int main(int argc, char * argv[]) {
  if (argc != 2) {
    Usage(argv[0]);
    return 1;
  }
  size_t n_elements;
  try {
    n_elements = stoll(argv[1]);
  } catch (exception & e) {
    cerr << "number-of-elements is not a valid number: " << e.what() << "\n";
    Usage(argv[0]);
    return 1;
  }
  Timer t;
  RunBenchmark(n_elements);
  t.Tock();
  std::cout << "\ntotal time: " << t.ElapsedTime() << " ms\n";
  return 0;
}
